let plain, property;

/* eslint-disable-next-line max-lines-per-function */
export default function arrayProxyMethods({prototype}, plane, prop) {
	plain = plane;
	property = prop;
	prototype.entries          = entries;
	prototype.keys             = keys;
	prototype.values           = values;
	prototype[Symbol.iterator] = values;
	Object.defineProperty(prototype, 'length', {get: length});

	prototype.at = at;

	const nonMod = [
		'concat', 'every', 'filter', 'find', 'findIndex', 'findLast',
		'findLastIndex', 'flat', 'flatMap', 'forEach', 'group', 'groupToMap',
		'includes', 'indexOf', 'join', 'lastIndexOf', 'map', 'reduce',
		'reduceRight', 'slice', 'some', 'toLocaleString', 'toString',
	];
	for(const key of nonMod) prototype[key] = noMod(key);

	for(const key of ['push', 'unshift']) prototype[key] = elAdder(key);
	for(const key of ['pop',    'shift']) prototype[key] = elRemover(key);

	prototype.copyWithin = copyWithin;
	prototype.fill       = fill;
	prototype.reverse    = reverse;
	prototype.sort       = sort;
	prototype.splice     = splice;
}

const privat = new WeakMap();
const priv = obj => privat.get(obj);
const arr  = obj => privat.get(obj).arr;

export function iniArrayProxy(that, el, path) {
	privat.set(that, {arr: [], refEl: el, path});
	Object.seal(that);
}

/* eslint no-invalid-this: "off" */
function* entries() {
	for(const [key, el] of arr(this).entries()) yield [key, el.get()];
}
function* keys() {for(const el of arr(this).keys()) yield el;}
function* values() {for(const el of arr(this)) yield el.get();}

function at(idx) {return arr(this).at(idx).get();}

function toEl({refEl, path}, el, idx) {
	const prop = property(refEl, [...path, idx]);
	prop.set(el);
	return prop;
}

// function mutator(f, name = f.name) {return function(...args) {
// 	const p = priv(this);
// 	return reduce(p.path, name, (...args) => f(...args), ...args);
// };}

function elAdder(key) {return function(...el) {
	const p = priv(this);
	p.arr[key](...el.map((el, idx) => toEl(p, el, p.arr.length + idx)));
	return p.arr.length;
};}
function elRemover(key) {return function() {
	const ret = arr(this)[key]();
	return ret && plain(ret);
};}


function noMod(key) {return function(...arg) {
	return plain(this)[key](...arg);
};}

function length() {return arr(this).length;}


function fill(value, start, end) {
	const p = priv(this);
	const {arr} = p;
	if(!arr.length) return this;
	start = (start < 0) ? arr.length + start : start || 0;
	end   = (end   < 0) ? arr.length +   end : (
		(end === undefined) ? arr.length : ((end > arr.length) ? arr.length : end)
	);
	for(let i = start; i < end; i++) arr[i] = toEl(p, value, i);
	return this;
}

function copyWithin(...arg) {
	const p = priv(this);
	const {arr} = p;
	arr.copyWithin(...arg);
	for(let i = 0; i < arr.length; i++) arr[i] = toEl(p, plain(arr[i]), i);
	return this;
}

function reverse() {
	const p = priv(this);
	const {arr} = p;
	arr.reverse();
	for(let i = 0; i < arr.length; i++) arr[i] = toEl(p, plain(arr[i]), i);
	return this;
}

/* eslint-disable-next-line max-lines-per-function */
function sort(func) {
	const p = priv(this);
	const {arr} = p;
	arr.sort((a, b) => {
		a = a.get();
		b = b.get();
		if(func)  return func(a, b);
		if(a < b) return -1;
		if(a > b) return  1;
		else      return  0;
	});
	for(let i = 0; i < arr.length; i++) arr[i] = toEl(p, plain(arr[i]), i);
	return this;
}

function splice(start, delCount, ...els) {
	const p = priv(this);
	if(start < 0) start = p.arr.length + start;
	if(delCount === undefined) delCount = p.arr.length;
	return p.arr.splice(start, delCount,
		...els.map((el, idx) => toEl(p, el, start + idx)),
	).map(el => plain(el));
}
